/**
 * @file application.h
 * @brief Contains the %Application class.
 */

#ifndef APPLICATION_H
#define APPLICATION_H

// System headers
#include <stdio.h>

// Passed dependencies
#include "logger.h"
#include "timermanager.h"
#include "inputpin.h"
#include "externalinterrupt.h"

// Aggregated members
#include "rc5receiver.h"
#include "rc5telegram.h"
#include "synchronizedringbuffer.h"

/**
 * @brief Main %Application class that aggregates all other components.
 *
 * This class contains the application specific logic and handles the received RC5 telegrams
 * by writing their content to the Logger. It has an internal buffer for a small number of
 * telegrams. Its main purpose is to decouple the interrupt context from the main context:
 * - Telegrams are received by onTelegram() in ISR context
 * - Telegrams are logged by run() in main context
 */

class Application : public RC5Receiver::TelegramHandler
{
public:
	/**
	 * The constructor stores the passed dependencies inside the member variables.
	 * @param[in] logger Logger instance to use
	 * @param[in] timerManager TimerManager instance to use
	 * @param[in] samplePin SamplePin instance passed to the RC5Receiver
	 * @param[in] extInterrupt ExternalInterrupt instance passed to the RC5Receiver
	 */
	Application(Logger *logger, TimerManager *timerManager, InputPin *samplePin, ExternalInterrupt *extInterrupt) :
		m_logger(logger),
		m_receiver(timerManager, samplePin, extInterrupt, this),
		m_inputOverflow(false)
	{
	}

	/**
	 * Initializes the application and its sub components.
	 */
	void init(void)
	{
		m_receiver.init();
		m_logger->write(LOG_INFO, "RC5 demo ready.");
	}

	/**
	 * Deinitializes the application and its sub components.
	 */
	void deinit(void)
	{
		m_receiver.deinit();
	}

	/**
	 * Method to be called from main context to perform lengthy operations.
	 * Here the telegrams are dequeued from the input queue and print using
	 * the logger.
	 * Input queue overflows are also logged.
	 */
	void run(void)
	{
		RC5Telegram telegram;
		while (m_inputQueue.dequeue(&telegram))
		{
			print(telegram);
		}

		if (m_inputOverflow)
		{
			m_logger->write(LOG_WARNING, "Telegram loss!");
			m_inputOverflow = false;
		}
	}

	/**
	 * Telegram handler that stores the passed telegram inside the input
	 * queue. In case the queue is full the overflow flag is set.
	 * @param[in] telegram Telegram to handle
	 */
	virtual void onTelegram(const RC5Telegram &telegram) override
	{
		if (m_inputQueue.enqueue(telegram) == false)
		{
			m_inputOverflow = true;
		}
	}

private:
	/**
	 * Prints the content of a RC5 telegram using the logger.
	 * @param[in] telegram Telegram to print
	 */
	void print(const RC5Telegram &telegram)
	{
		char buffer[50];

		sprintf(buffer, "Field: %i, Toggle: %i, Address: %i, Code: %i",
			telegram.getFieldBit(), telegram.getToggleBit(),
			telegram.getAddress(), telegram.getCode());

		m_logger->write(LOG_INFO, buffer);
	}

	/// Reference to Logger instance
	Logger *m_logger;

	/// Instance of RC5Receiver component
	RC5Receiver m_receiver;

	/// Input queue to decouple ISR context from main context
	SynchronizedRingBuffer<RC5Telegram, 3> m_inputQueue;

	/// Overflow flag indicating telgram loss in ISR context
	bool m_inputOverflow;
};

#endif
