/**
 * @file bitbuffer.h
 * @brief Contains the %BitBuffer class.
 */

#ifndef BITBUFFER_H
#define BITBUFFER_H

#include "bit.h"

/**
 * Template class to store a series of bits.
 * @param BASETYPE Integer base type to use as internal buffer
 * @param MAXBITS Maximum number of bits to store
 */
template <typename BASETYPE, uint8_t MAXBITS>
class BitBuffer
{
public:
	BitBuffer(void) : m_buffer(0), m_numBits(0)
	{
	}

	/**
	 * Clears the internal buffer.
	 */
	void clear(void)
	{
		m_buffer = 0;
		m_numBits = 0;
	}

	/**
	 * Adds a bit to the least significant position, shifting
	 * existing bits to the left.
	 * @note When full, new bits are silently discarded.
	 * @param[in] bit Bit to add
	 * @return True if bit was added, false if buffer is full.
	 */
	bool add(Bit bit)
	{
		bool added = false;

		if (m_numBits < MAXBITS)
		{
			m_buffer = (m_buffer << 1) | static_cast<uint8_t>(bit);
			m_numBits++;
			added = true;
		}

		return added;
	}

	/**
	 * Returns true if buffer is full.
	 * @return True if buffer is full, false otherwise.
	 */
	bool isFull(void) const
	{
		return m_numBits == MAXBITS;
	}

	/**
	 * Returns true if buffer is empty.
	 * @return True if buffer is empty, false otherwise.
	 */
	bool isEmpty(void) const
	{
		return m_numBits == 0;
	}

	/**
	 * Gets the content of the internal buffer.
	 */
	BASETYPE get(void) const
	{
		return m_buffer;
	}

private:
	/// Internal buffer to hold bits.
	BASETYPE m_buffer;

	/// Current number of bits in buffer.
	uint8_t m_numBits;
};

#endif
