/**
 * @file bitdetector.h
 * @brief Contains the %BitDetector class and the BitHandler interface.
 */

#ifndef BITDETECTOR_H
#define BITDETECTOR_H

// Base classes
#include "externalinterrupt.h"
#include "eventreceiver.h"

// Interface types
#include "bit.h"
#include "timermanager.h"
#include "inputpin.h"
#include "oneshottimer.h"

/**
 * @brief Class that detects bits using the manchester coding.
 *
 * The BitDetector uses an InputPin and an ExternalInterrupt instance to detect signal edges
 * that encode a bit using the manchester coding. The procedure is:
 * - Wait for interrupt in the middle of the bit (falling edge => ONE bit, rising edge => ZERO bit)
 * - Start timeout for 3/4 of the length of a bit
 * - Sample input level to decide which interrupt will occur next (rising or falling edge)
 *
 * This also forms the internal state machine:
 * @startuml
 * !include bitdetector-states.pu
 * @enduml
 */
class BitDetector : public ExternalInterrupt::InterruptHandler, public EventReceiver
{
public:
	/**
	 * Interface for components that should be able to handle bits.
	 */
	class BitHandler
	{
	public:
		/**
		 * Handler called for each detected bit.
		 * @param[in] bit Detected bit
		 */
		virtual void onBitDetected(Bit bit) = 0;
	};

	/**
	 * Constructor.
	 * @param[in] timerManager TimerManager instance to use
	 * @param[in] samplePin InputPin connected to TSOP IC
	 * @param[in] extInterrupt ExternalInterrupt instance to manage interrupts on samplePin line
	 * @param[in] handler Handler that receives the detected bits
	 */
	BitDetector(TimerManager *timerManager, InputPin *samplePin, ExternalInterrupt *extInterrupt, BitHandler *handler) :
		m_timerManager(timerManager), m_samplePin(samplePin), m_extInterrupt(extInterrupt), m_handler(handler),
		m_sampleTimeout(this, SAMPLE_TIMEOUT_ELAPSED, 5), // 1250us timeout @ 250us tick
		m_state(NOT_INITIALIZED)
	{
		m_extInterrupt->setInterruptHandler(this);
	}

	/**
	 * Destructor to make sure that interrupt handler is unregistered.
	 */
	~BitDetector(void)
	{
		m_extInterrupt->setInterruptHandler(nullptr);
	}

	/**
	 * Initialization.
	 * Starts with waiting for a ONE start bit.
	 */
	void init(void)
	{
		m_timerManager->add(&m_sampleTimeout);
		setState(WAIT_FOR_FALLING_EDGE); // ONE start bit
	}

	/**
	 * Deinitialization.
	 * Enters NOT_INITIALIZED state and disables interrupt.
	 */
	void deinit(void)
	{
		m_timerManager->remove(&m_sampleTimeout);
		m_extInterrupt->enableInterrupt(false);
		setState(NOT_INITIALIZED);
	}

	/**
	 * Handler for external interrupts.
	 */
	virtual void onExternalInterrupt(void) override
	{
		switch (m_state)
		{
		case WAIT_FOR_FALLING_EDGE:
			m_handler->onBitDetected(ONE);
			setState(WAIT_FOR_SAMPLE);
			break;
		case WAIT_FOR_RISING_EDGE:
			m_handler->onBitDetected(ZERO);
			setState(WAIT_FOR_SAMPLE);
			break;
		default:
			break;
		}
	}

	/**
	 * Event handler.
	 * Receives the notification event when the timeout elapsed.
	 * Implements state transition back to either WAIT_FOR_FALLING_EDGE
	 * or WAIT_FOR_RISING_EDGE depending on current input level.
	 * @param[in] event Event to handle
	 */
	virtual void onEvent(const Event &event) override
	{
		if (m_state == WAIT_FOR_SAMPLE && event.getId() == SAMPLE_TIMEOUT_ELAPSED)
		{
			setState(m_samplePin->getValue() ? WAIT_FOR_FALLING_EDGE : WAIT_FOR_RISING_EDGE);
		}
	}

private:
	/// Events used by this component
	enum EventIds
	{
		SAMPLE_TIMEOUT_ELAPSED ///< Event for elapsed sample timeout
	};

	/// States of internal state machine
	enum State
	{
		NOT_INITIALIZED, ///< Not initialized
		WAIT_FOR_FALLING_EDGE, ///< Waiting for falling edge
		WAIT_FOR_RISING_EDGE, ///< Waiting for rising edge
		WAIT_FOR_SAMPLE, ///< Waiting for timeout to sample input
	};

	/**
	 * Sets internal state machine state.
	 * Implements the On Entry actions.
	 * @param[in] state State to set
	 */
	void setState(State state)
	{
		m_state = state;

		switch (m_state)
		{
		case NOT_INITIALIZED:
			break;
		case WAIT_FOR_FALLING_EDGE:
			m_extInterrupt->setTriggerMode(ExternalInterrupt::FALLING_EDGE);
			m_sampleTimeout.stop();
			m_extInterrupt->enableInterrupt(true);
			break;
		case WAIT_FOR_RISING_EDGE:
			m_extInterrupt->setTriggerMode(ExternalInterrupt::RISING_EDGE);
			m_sampleTimeout.stop();
			m_extInterrupt->enableInterrupt(true);
			break;
		case WAIT_FOR_SAMPLE:
			m_extInterrupt->enableInterrupt(false);
			m_sampleTimeout.restart();
			break;
		}
	}

	/// Reference to TimerManager
	TimerManager *m_timerManager;

	/// Reference to pin to sample
	InputPin *m_samplePin;

	/// Reference to interrupt control object
	ExternalInterrupt *m_extInterrupt;

	/// Reference to handler for detected bits
	BitHandler *m_handler;

	/// Timer to monitor telegram timeout
	OneShotTimer<uint8_t> m_sampleTimeout;

	/// State of internal state machine
	State m_state;
};

#endif
