/**
 * @file logger.h
 * @brief Contains the %Logger class and related types.
 */

#ifndef LOGGER_H
#define LOGGER_H

// Device to use for output
#include "bufferedusart.h"

/**
 * @brief Importance level of messages to log.
 *
 * Each level contains all previous levels, e.g. when the current
 * output threshold is set to INFO, messages with level WARNING are
 * also logged, but messages with level DEBUG are discarded.
 */
enum Level
{
	LOG_ERROR, ///< Error messages
	LOG_WARNING, ///< Warning messages
	LOG_INFO, ///< General information messages
	LOG_DEBUG, ///< Messages for debugging purposes
};

/**
 * @brief Logger class that takes messages and forwards them to the configured USART.
 *
 * The logger doesn't store any information, all messages are directly passed.
 * It thus serves as an interface, also providing a filter mechanism based
 * on message importance levels (see #Level).
 */
class Logger
{
public:
	/**
	 * Constructor.
	 * @param[in] usart Configured USART device to use for output
	 * @param[in] level Initial threshold for filtering (see #Level).
	 */
	Logger(Usart *usart, Level level = LOG_INFO);
	~Logger(void);

	/**
	 * Writes a message of the specified importance level.
	 *
	 * This applies the filtering (see #Level) and prepends the
	 * importance level as string (e.g. `"[warn] text"`).
	 * @note Line endings are added automatically.
	 * @param[in] level Importance level of message
	 * @param[in] s Message to write
	 */
	void write(Level level, const char *s);

	/**
	 * Sets the importance level threshold for filtering.
	 * @param[in] level Threshold to set.
	 */
	void setLevel(Level level);

private:
	/**
	 * Writes a string to the associated UART device.
	 * @param[in] s String to write
	 */
	void write(const char *s);

	/// Device to use for writing data asynchronously.
	BufferedUsart<0, 100> m_usart;

	/// Current threshold for filtering.
	Level m_level;
};

#endif
