/**
 * @file main.cpp
 * @brief Main file which creates all dependencies to concrete hardware.
 *
 * In this file all instances are created that have a binding to a special
 * hardware resource (e.g. depending on the pin layout) or require
 * configuration that shouldn't be scattered throughout the entire program
 * (e.g. timer configuration).
 *
 * The benefit of this approach is that all dependencies are located in one
 * place and thus are easy to spot. Additionally the other components are
 * fairly generic and don't require to now specific details (e.g. that the
 * input pin is located on a GPIO port).
 *
 * The downside obviously is that this approach requires to pass lots of
 * dependencies down into the nested components, which means that each
 * component needs to know every dependency of all of its children.
 */

#include "usart0.h"
#include "timer0.h"
#include "portd.h"
#include "gpioportpin.h"
#include "int0.h"
#include "globalinterruptflag.h"

#include "logger.h"
#include "timermanager.h"
#include "application.h"

/**
 * @brief Entry point
 *
 * Instances are created as local objects on the stack.
 * Temporary variables are wrapped inside their own scope so that they are destroyed immediately after use.
 * @note This function never returns.
 */
int main(void)
{
	Usart0 usart;
	{
		Usart::Config config;
		Usart::autoFill<8000000UL>(config, 9600, "8N1");
		usart.configure(config);
	}
	Logger logger(&usart, LOG_DEBUG);
	logger.write(LOG_DEBUG, "Logger created.");

	Timer0 timer;
	{
		Timer::Config config;
		Timer::autoFill<8000000UL>(config, 250);
		timer.configure(config);
	}
	TimerManager timerManager(&timer);

	PortD port;
	GpioPortPin RC5Pin(&port, 2);
	RC5Pin.setDirection(INPUT);
	RC5Pin.setPullup(true);
	Int0 RC5Interrupt;

	Application app(&logger, &timerManager, &RC5Pin, &RC5Interrupt);
	app.init();

	GlobalInterruptFlag::set();
	for (;;)
	{
		app.run();
	}
	GlobalInterruptFlag::clear();

	app.deinit();
	return 0;
}
