/**
 * @file rc5receiver.h
 * @brief Contains the %RC5Receiver class and the TelegramHandler interface.
 */

#ifndef RC5RECEIVER_H
#define RC5RECEIVER_H

// Base classes
#include "bitdetector.h"
#include "eventreceiver.h"

// Interface types
#include "rc5telegram.h"
#include "timermanager.h"
#include "inputpin.h"
#include "externalinterrupt.h"

// Aggregated members
#include "bitbuffer.h"
#include "oneshottimer.h"

/**
 * @brief Class that accumulates individual bits into RC5 telegrams.
 *
 * The %RC5Receiver implements the BitHandler interface provided by the BitDetector.
 * It takes the bits and stores them inside a BitBuffer until a set of 14 bits forming
 * a RC5 telegram is complete. Then it is wrapped inside a RC5Telegram instance and
 * passed via the TelegramHandler interface.
 *
 * When the first bit of a telegram is detected, a 30ms timer is started. In case this
 * timer elapses, reception of the current RC5 telegram is discarded and the BitBuffer
 * is cleared (a regular telegram should not exceed 24,889 ms in length). This makes
 * the application more robust against faulty transmissions and noise on the signal line.
 *
 * The %RC5Receiver implements this state machine:
 * @startuml
 * !include rc5receiver-states.pu
 * @enduml
 */
class RC5Receiver : public BitDetector::BitHandler, public EventReceiver
{
public:
	/**
	 * Interface for components that should be able to handle RC5 telegrams.
	 */
	class TelegramHandler
	{
	public:
		/**
		 * Telegram handler called for each received RC5 telegram.
		 * @param[in] telegram Received telegram
		 */
		virtual void onTelegram(const RC5Telegram &telegram) = 0;
	};

	/**
	 * Constructor. Most of the parameters are passed to the BitDetector member.
	 * @param[in] timerManager TimerManager instance to use
	 * @param[in] samplePin InputPin connected to TSOP IC
	 * @param[in] extInterrupt ExternalInterrupt instance to manage interrupts on samplePin line
	 * @param[in] handler Handler that receives the accumulated RC5 telegrams
	 */
	RC5Receiver(TimerManager *timerManager, InputPin *samplePin, ExternalInterrupt *extInterrupt, TelegramHandler *handler) :
		m_timerManager(timerManager),
		m_handler(handler),
		m_detector(timerManager, samplePin, extInterrupt, this),
		m_telegramTimeout(this, TELEGRAM_TIMEOUT_ELAPSED, 120), // 30ms timeout @ 250us tick
		m_state(NOT_INITIALIZED)
	{
	}
	
	/**
	 * Initialization.
	 * Enters WAIT_FOR_START_BIT state and initializes BitDetector.
	 */
	void init(void)
	{
		m_timerManager->add(&m_telegramTimeout);
		setState(WAIT_FOR_START_BIT);
		m_detector.init();
	}

	/**
	 * Deinitialization.
	 * Enters NOT_INITIALIZED state and deinitializes BitDetector.
	 */
	void deinit(void)
	{
		m_detector.deinit();
		m_timerManager->remove(&m_telegramTimeout);
		setState(NOT_INITIALIZED);
	}

	/**
	 * Handler for detected bits.
	 * Implements the state transitions triggered by detected bits.
	 * @param[in] bit Detected bit
	 */
	virtual void onBitDetected(Bit bit) override
	{
		switch (m_state)
		{
		case NOT_INITIALIZED:
			break;
		case WAIT_FOR_START_BIT:
			if (bit == ONE)
			{
				m_buffer.add(bit);
				setState(RECEIVE_BIT_STREAM);
			}
			break;
		case RECEIVE_BIT_STREAM:
			m_buffer.add(bit);
			if (m_buffer.isFull())
			{
				m_handler->onTelegram(m_buffer.get());
				setState(WAIT_FOR_START_BIT);
			}
			break;
		}
	}

	/**
	 * Event handler.
	 * Receives the notification event when the timeout elapsed.
	 * Implements state transition back to WAIT_FOR_START_BIT.
	 * @param[in] event Event to handle
	 */
	virtual void onEvent(const Event &event) override
	{
		if (event.getId() == TELEGRAM_TIMEOUT_ELAPSED)
		{
			setState(WAIT_FOR_START_BIT);
		}
	}

private:
	/// Events used by this component
	enum EventIds
	{
		TELEGRAM_TIMEOUT_ELAPSED, ///< Event for elapsed telegram timeout
	};

	/// States of internal state machine
	enum State
	{
		NOT_INITIALIZED, ///< Not initialized
		WAIT_FOR_START_BIT, ///< Waiting for start bit (1)
		RECEIVE_BIT_STREAM, ///< Receiving remaining bits of telegram
	};

	/**
	 * Sets internal state machine state.
	 * Implements the On Entry actions.
	 * @param[in] state State to set
	 */
	void setState(State state)
	{
		m_state = state;

		switch (m_state)
		{
		case NOT_INITIALIZED:
			break;
		case WAIT_FOR_START_BIT:
			m_buffer.clear();
			m_telegramTimeout.stop();
			break;
		case RECEIVE_BIT_STREAM:
			m_telegramTimeout.restart();
			break;
		}
	}

	/// Reference to TimerManager
	TimerManager *m_timerManager;

	/// Reference to TelegramHandler
	TelegramHandler *m_handler;

	/// BitDetector instance
	BitDetector m_detector;

	/// Internal buffer to accumulate bits in telegram
	BitBuffer<uint16_t,14> m_buffer;

	/// Timer to monitor telegram timeout
	OneShotTimer<uint8_t> m_telegramTimeout;

	/// State of internal state machine
	State m_state;
};

#endif
