/**
 * @file rc5telegram.h
 * @brief Contains the %RC5Telegram class.
 */

#ifndef RC5TELEGRAM_H
#define RC5TELEGRAM_H

// For basic types
#include <stdint.h>
#include "bit.h"

/**
 * @brief Class representing a RC5 telegram.
 *
 * Internally it stores data in a bitfield that matches the binary
 * layout of the RC5 protocol data units. It provides getters and
 * setters for both, a raw bit pattern (as uint16_t) and individual
 * elements (e.g. address or code). Thus it can be also used for
 * preparing a RC5 telegram for transmission (even though this isn't
 * used in this demo).
 */
class RC5Telegram
{
public:
	/**
	 * Constructor that optionally initializes from a bit pattern.
	 * @param[in] bits Pattern to initialize with
	 */
	RC5Telegram(uint16_t bits = 0)
	{
		m_pdu.value = bits;
	}

	/**
	 * Cast operator for converting telegram into bit pattern.
	 * @note The start bit is implicitly set to 1.
	 */
	operator uint16_t() const
	{
		ProtocolDataUnit tmp = m_pdu;

		tmp.bits.unused = 0;
		tmp.bits.start = 1;

		return tmp.value;
	}

	/**
	 * Getter for field bit.
	 * @return Field bit
	 */
	Bit getFieldBit(void) const
	{
		return static_cast<Bit>(m_pdu.bits.field);
	}

	/**
	 * Setter for field bit.
	 * @param[in] state State of field bit to set
	 */
	void setFieldBit(Bit state)
	{
		m_pdu.bits.field = state;
	}

	/**
	 * Getter for toggle bit.
	 * @return Toggle bit
	 */
	Bit getToggleBit(void) const
	{
		return static_cast<Bit>(m_pdu.bits.toggle);
	}

	/**
	 * Setter for toggle bit.
	 * @param[in] state State of toggle bit to set
	 */
	void setToggleBit(Bit state)
	{
		m_pdu.bits.toggle = state;
	}

	/**
	 * Getter for address.
	 * @return Address
	 */
	uint8_t getAddress(void) const
	{
		return m_pdu.bits.address;
	}

	/**
	 * Setter for address.
	 * @param[in] address Address to set
	 */
	void setAddress(uint8_t address)
	{
		m_pdu.bits.address = address;
	}

	/**
	 * Getter for code.
	 * @return Code
	 */
	uint8_t getCode(void) const
	{
		return m_pdu.bits.code;
	}

	/**
	 * Setter for code.
	 * @param[in] code Code to set
	 */
	void setCode(uint8_t code)
	{
		m_pdu.bits.code = code;
	}

private:
	/**
	 * Type to store a protocol data unit as bitfield and as
	 * bit pattern (unsigned integer value).
	 */
	union ProtocolDataUnit
	{
		/// Bitfield portion
		struct {
			uint16_t code    : 6; ///< Code (6 bit)
			uint16_t address : 5; ///< Address (5 bit)
			uint16_t toggle  : 1; ///< Toggle bit (1 bit)
			uint16_t field   : 1; ///< Field bit (1 bit)
			uint16_t start   : 1; ///< Start bit (1 bit)
			uint16_t unused  : 2; ///< Unused bits (2 bits)
		} bits;
		uint16_t value; ///< 16 bit value with raw bit pattern
	};

	/// Internal buffer holding the protocol data unit
	ProtocolDataUnit m_pdu;
};

#endif
